#!/usr/bin/env python
#*****************************************************************************#
#**
#**  Realtime Annotation Tool (Rat)
#**
#**    Kyle D. Feuz, 2012
#**
#** Tools by the Center for Advanced Studies in Adaptive Systems at
#**  the School of Electrical Engineering and Computer Science at
#**  Washington State University
#** 
#** Copyright Washington State University, 2012
#** Copyright Kyle D. Feuz, 2012
#** 
#** Licenced under the Creative Commons (http://creativecommons.org):
#** Attribution-NonCommercial-ShareAlike 3.0 Unported
#** 
#** If this code is used for public research, any resulting publications need
#** to cite work done by Kyle D. Feuz at the Center for Advanced Study of 
#** Adaptive Systems (CASAS) at Washington State University.
#** 
#** Contact: Kyle D. Feuz (kyle.feuz@wsu.edu)
#** Contact: Diane J. Cook (cook@eecs.wsu.edu)
#*****************************************************************************#

import optparse
import os
import signal
import time
import re
import random
import ConfigParser
import pygtk
import threading
pygtk.require('2.0')
import gtk, gobject
try:
   import xml.etree.ElementTree as ET
except ImportError:
   import cElementTree as ET
try:
   from collections import OrderedDict
except ImportError:
   from OrderedDict import OrderedDict
try:
   import fcntl
except ImportError:
   print "Warning unable to import fcntl"
   print "Do not run more than one instance of the RAT!"
try:
   import pgdb
except ImportError:
   print "Warning pgdb module not loaded"
from datetime import datetime
from casas import xmpp


NAME = "RAT"
FULL_NAME = "Realtime Annotation Tool"
VERSION = "1.0"

class Annotator:
   """
   This class handles all of the communication with the middleware
   to annotate the data
   """

   def __init__(self, parent, user, passw, man, local):
      self.name = NAME
      self.version = "v" + VERSION
      self.parent = parent
      self.username = user
      self.password = passw
      self.manager = man
      self.experiment = ""
      self.dataset = ""
      self.open_events = {}
      self.event_id = 0
      self.cid = 0
      self.connected = False
      if local:
         self.xmpp = xmpp.LocalConnection(self.name)
      else:
         self.xmpp = xmpp.Connection(self.name)
      self.xmpp.set_authd_callback(self.connect_to_manager)
      if self.manager != "":
         self.xmpp.set_custom_manager(self.manager)
      return

   def connect(self):
      """
      Connect to the middleware
      """
      self.xmpp.connect(self.username, self.password)
      return

   def connect_to_manager(self):
     """
     Called after authenticated to XMPP. Registers channels
     """
     self.connected = True
     print self.name, "Connected"
     self.xmpp.log(self.name + " connected")
     self.xmpp.publish_channel("rawevents")
     self.xmpp.publish_channel("control")
     self.xmpp.publish_channel("tag")
     self.xmpp.set_channel_callback("rawevents", self.handle_raw_events)
     self.xmpp.subscribe_channel("rawevents")
     self.xmpp.set_channel_callback("control", self.handle_control_events)
     self.xmpp.subscribe_channel("control")
     self.xmpp.set_channel_callback("tag", self.handle_tagged_events)
     self.xmpp.subscribe_channel("tag")
     self.parent.connected_to_middleware()
     return


   def set_dataset(self, dataset):
      """
      Set the current dataset name to be used in the db
      """
      self.dataset = dataset

   def set_experiment(self, experiment):
      """
      Set the current experiment name to be used in the db
      """
      self.experiment = experiment

   def publish_data(self, data, tags):
      """
      Publish events generated by the GUI
      """
      msg = "<event><by>%s</by>" % self.name
      msg += "<type>%s</type>" % self.name
      msg += "<serial>%s</serial>" % self.name+self.version
      msg += "<location>system</location>"
      msg += "<message>%s</message>" % str(data)
      msg += "<category>entity</category>"
      msg += "<event_id>%i</event_id>" % self.event_id
      msg += "</event>"
      self.open_events[self.event_id] = tags
      self.event_id += 1
      self.event_id %= 10000
      self.xmpp.publish(msg)

   def publish_command(self, value, serial="", location="", command="prompt"):
      """
      Publish command to the command channel
      """
      self.xmpp.publish_command(self.name, command, value, serial, location, cid=self.cid)
      self.cid += 1
      self.cid %= 10000


   def tag_event(self, event, tag):
      """
      Tag an event
      """
      self.xmpp.tag_event(self.name, self.experiment, self.dataset, tag.label, tag.value, event.by, event.mtype, event.serial, event.location, event.message, event.category, event.epoch, event.uuid) 
   
   def handle_raw_events(self, msg):
      """
      Check incoming events for events currently needing tagged and tag them
      """
      if re.search("<publish>.*?</publish>", str(msg)):
         try:
            eid = int(xmpp.getData("event_id", msg))
         except:
            eid = -1
         if eid in self.open_events:
            tags = self.open_events[eid]
            del self.open_events[eid]
         else:
            tags = [Tag()]
         event = xmpp.Event(msg)
         event.epoch = xmpp.getData("epoch", msg)
         for tag in tags:
            self.tag_event(event, tag)
   
   def handle_tagged_events(self, msg):
      """
      Notify GUI of received tagged events
      """
      if re.search("<publish>.*?</publish>", str(msg)):
         by = xmpp.getData("by", msg)
         if self.name == by:
            t = xmpp.Tag(msg)
            t.epoch = xmpp.getData("epoch", msg)
            t.isTag = True
            self.parent.show_output(t)


   def handle_control_events(self, msg):
      """
      Notify GUI of received control events
      """
      if re.search("<publish>.*?</publish>", str(msg)):
         c = xmpp.Command(msg)
         by = xmpp.getData("by", msg)
         if self.name == by:
            c.epoch = xmpp.getData("epoch", msg)
            c.isTag = False
            self.parent.show_output(c)
         event = xmpp.Event(msg);
         event.message = "%s|%s|%s|%s" % (c.command, c.value, c.replyto, c.cid)
         event.epoch = xmpp.getData("epoch", msg)
         event.serial = c.serial
         event.location = c.location
         event.by = c.by
         event.category = 'control'
         event.mtype = 'control'
         self.tag_event(event, Tag())

   def finished(self):
      """Called after everything is cleaned up"""
      print "Disconnecting"
      self.xmpp.disconnect()
      self.connected = False

   def finish(self):
      """
      Cleans up when we are done (exiting).
      """
      print "Cleaning up"
      self.xmpp.unpublish_channel("rawevents")
      self.xmpp.unpublish_channel("control")
      self.xmpp.unsubscribe_channel("rawevents")
      self.xmpp.log(self.name + " disconnecting")
      self.xmpp.callLater(0, self.finished)



# Update the value of the progress bar so that we get
# some movement
def progress_timeout(pbar):
    pbar.pulse()
    return True

class RAT_GUI:

   #####################
   #                   #
   #   GUI callbacks   #
   #                   #
   #####################
   def cb_questionnaire_button(self, widget, data):
      self.show_questionnaire(data)

   def cb_task_button(self, widget, data):
      self.move_task(data)

   def cb_event_button(self, widget, data):
      exp = data[0]
      task = data[1]
      event = data[2]
      other = data[3]
      widget.modify_bg(gtk.STATE_NORMAL, self.gray)
      if event.type == Event.BUTTON:
         values = [""]
      elif event.type == Event.FIELD:
         values = []
         for field in other:
            values.append(field.get_text())
      elif event.type == Event.CHOICE:
         values = []
         for cmenu in other:
            index = cmenu.get_history()
            values.append(event.choices[index])
      msg = "[%s:%i][%s:%i][%s:%i]" % (exp.name, exp.id, task.name, task.id, event.name, event.id)
      tags = []
      for value in values:
         tag = Tag()
         tag.label = event.name
         tag.value = value
         tags.append(tag)
      self.annotator.publish_data(msg, tags)

   def cb_startstop_task(self, widget, data):
      exp = data[0]
      task = data[1]
      text = widget.get_label()
      if text == "START":
         widget.set_label("STOP")
         widget.modify_bg(gtk.STATE_NORMAL, self.red)
         if task is not None and task.prompt is not None:
            self.cb_update_prompt(widget, task.prompt)
      elif text == "STOP":
         widget.set_label("START")
         widget.modify_bg(gtk.STATE_NORMAL, self.yellow)
      tag = Tag()
      if task != None:
         msg = "[%s:%i][%s:%i]" % (exp.name, exp.id, task.name, task.id)
         tag.label = task.name
      else:
         msg = "[%s:%i]" % (exp.name, exp.id)
         tag.label = exp.name
      tag.value = text
      self.annotator.publish_data(msg, [tag])

   def cb_update_prompt(self, widget, data):
      if self.prompt != None:
         index = self.prompt.level-1
         if self.prompt.played:
            color = self.gray
         else:
            color = self.lgreen
         self.prompt.buttons[index].modify_bg(gtk.STATE_NORMAL, color)
      self.prompt = data
      if self.prompt.level > 0:
         index = self.prompt.level-1
         self.prompt.buttons[index].modify_bg(gtk.STATE_NORMAL, self.green)
      else:
         self.prompt = None

   def cb_quick_prompt(self, group, widget, key, modifier):
      if self.prompt != None:
         data = "%i%s" % (self.prompt.level, self.prompt.filename)
         index = self.prompt.level-1
         self.prompt.buttons[index].modify_bg(gtk.STATE_NORMAL, self.dgreen)
         self.prompt.played = True
         for device in self.prompt_devices:
            if device.name == "_Default" or device.name == "Default":
               for did in self.prompt.default_devices:
                  d = self.study.actual_devices[int(did)]
                  if d in self.prompt_devices:
                     continue
                  else:
                     self.annotator.publish_command(data, d.serial, d.location)
            else:
               self.annotator.publish_command(data, device.serial, device.location)

   def cb_prompt(self, widget, data):
      level = data[0]
      prompt = data[1]
      value = level + prompt.filename
      widget.modify_bg(gtk.STATE_NORMAL, self.gray)
      for device in self.prompt_devices:
         if device.name == "_Default" or device.name == "Default":
            for did in prompt.default_devices:
               d = self.study.actual_devices[int(did)]
               if d in self.prompt_devices:
                  continue
               else:
                  self.annotator.publish_command(value, d.serial, d.location)
         else:
            self.annotator.publish_command(value, device.serial, device.location)

   def cb_reconnect(self, widget, data=None):
      self.connect_dialog.show()

   def cb_connect_to_middleware(self, widget, data=None):
      self.pbar.set_fraction(0.0)
      self.pbar.show()
      username = self.cd_userfield.get_text()
      password = self.cd_passfield.get_text()
      manager = self.cd_managerfield.get_text()
      local = self.cd_local.get_active()
      self.annotator = Annotator(self, username, password, manager, local) 
      self.annotator.connect()

   def cb_set_participant(self, widget, data):
      error = False
      if self.pd_new_rbutton.get_active():        #New PSID
         psid = self.pd_new_entry.get_text().strip()
         #Check for exsiting
         if psid in self.psid_list:
            error = True
            msg = "The new PSID is already used"
         #Check for blank psid
         elif psid == "":
            error = True
            msg = "The PSID cannot be blank"
         #Log new PSID info if no errors occurred
         else:
            self.clear_prompt_devices()
            index = self.pd_prompt_omenu.get_history()
            index = self.pd_prompt_omenu_translation[index]
            self.set_prompt_device(index)
            index = self.pd_condition_omenu.get_history()
            index = self.pd_condition_omenu_translation[index]
            self.condition = self.study.conditions[index]
            self.seed = self.seed_generator.random()
            mesg = "[New Participant:%s]" % (psid)
            tags = []
            tag = Tag()
            tag.label = "New Participant"
            tag.value = psid
            tags.append(tag)
            tag = Tag()
            tag.label = "prompt"
            tag.value = self.prompt_device.name.replace("_", "")
            tags.append(tag)
            tag = Tag()
            tag.label = "6 activity condition"
            tag.value = self.condition.name
            tags.append(tag)
            tag = Tag()
            tag.label = "seed"
            tag.value = self.seed
            tags.append(tag)
            self.pdata = (mesg, tags)

      elif self.pd_existing_rbutton.get_active():  #Existing PSID
         index = self.pd_existing_omenu.get_history()
         psid = data[index]
         # Use saved settings
         # Condition
         try:
            self.condition = self.study.conditions[self.psid_list[psid][0]]
         except KeyError:
            pass
         # Prompt Device
         try:
            self.clear_prompt_devices()
            self.set_prompt_device(self.psid_list[psid][1])
         except KeyError:
            pass
         self.seed = self.psid_list[psid][2]
         self.pdata = None

      else:                                        #TEST ID
         psid = self.testid
         self.seed = self.seed_generator.random()
         self.pdata = None
         
      if error:
         error_msg = gtk.MessageDialog(self.window, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, msg)
         error_msg.run()
         error_msg.destroy()
      else:
         root = self.xml_tree.getroot()
         for condition in root.findall("Condition"):
            if self.condition.id == int(condition.get("id", "-1")):
               self.condition = None
               self.load_condition(condition)
         self.cb_stop_participant(None, None)
         self.participant_dialog.hide()
         self.annotator.set_dataset(self.testid)
         self.make_inactive()
         self.psid = psid
         self.psid_label.set_text(psid)
         self.start_button.set_label("Start Participant: " + self.psid)

   def cb_start_participant(self, widget, data):
      self.load_experiment(self.eid)
      self.start_button.hide()
      self.stop_icon.hide()
      self.run_icon.show()
      self.annotator.set_dataset(self.psid)
      if self.pdata != None:
         self.annotator.publish_data(*self.pdata)
         self.pdata = None
      self.set_psid_used(self.psid, self.condition, self.prompt_device, self.seed)
      self.make_active()
   
   def cb_stop_participant(self, widget, data):
      self.load_experiment(self.eid)
      self.start_button.show()
      self.stop_icon.show()
      self.run_icon.hide()
      self.annotator.set_dataset(self.testid)
      self.make_inactive()


   def cb_pd_existing_omenu(self, widget, data=None):
      self.pd_existing_rbutton.set_active(True)
      self.pd_existing_rbutton.grab_focus()

   def cb_pd_new_entry(self, widget, event, data=None):
      self.pd_new_rbutton.set_active(True)

   def cb_mchange(self, widget, data=None):
      self.pd_ok_button.disconnect(self.pd_button_h)
      menu = gtk.Menu()
      for psid in self.psid_list:
         item = gtk.MenuItem(psid)
         item.show()
         menu.append(item)
      self.pd_existing_omenu.set_menu(menu)
      self.pd_button_h = self.pd_ok_button.connect('clicked', self.cb_set_participant, self.psid_list.keys())
      self.participant_dialog.show()

   def cb_mhistory(self, action, widget):
      if widget.active:
         self.history_dialog.show()
      else:
         self.history_dialog.hide()
      
   def cb_mhelp(self, widget, data=None):
      pass

   def cb_mabout(self, widget, data=None):
      aboutWindow = gtk.AboutDialog()
      aboutWindow.set_name("%s %s" %(FULL_NAME, VERSION))
      aboutWindow.set_authors(["Kyle D. Feuz","  Washington State University"])

      aboutWindow.connect("close", self.cb_close_about)
      aboutWindow.connect("response", self.cb_close_about)
      aboutWindow.run()

   def cb_mdevice(self, action, widget):
      if widget.active:
         self.prompt_devices.append(self.study.actual_devices[action])
      else:
         self.prompt_devices.remove(self.study.actual_devices[action])

   def cb_mexperiment(self, action, widget):
      if widget.get_active():
         self.load_experiment(action)

   def cb_quit(self, widget, event, data=None):
      if self.annotator != None and self.annotator.connected:
         self.annotator.finish()
      if self.timer != 0:
         gobject.source_remove(self.timer)
         self.timer = 0
      if data != None:
         print "Killing", data
         os.kill(data, signal.SIGKILL)
      print "Quiting"
      gtk.main_quit()
      return False


   def cb_changed_retag_omenu(self, widget, tv):
      choice = self.retag_choices[widget.get_history()]
      if choice == "Timing":
         tb = tv.get_buffer()
         text = tb.get_text(tb.get_start_iter(), tb.get_end_iter())
         if text.strip() == "":
            text = str(datetime.now())
            tb.set_text(text)
      return

   def cb_retag(self, widget, data):
      event, tv, omenu, d = data
      tag = Tag()
      tag.label = self.retag_choices[omenu.get_history()]
      tb = tv.get_buffer()
      text = tb.get_text(tb.get_start_iter(), tb.get_end_iter())
      tag.value = text.replace('\n', ' ')
      print "Retagging", event, tag.label, tag.value
      self.annotator.tag_event(event, tag)
      d.destroy()      

   def cb_history_row_activated(self, widget, row, col):
      model = widget.get_model()
      if len(row) > 1:
         return
      event = model[row][-1]
      title = "%s - %s" % (model[row][2], model[row][0])
      dialog = gtk.Dialog(title, self.history_dialog)
      dialog.connect("delete_event", self.cb_close_retag)
      dialog.set_default_size(300, 100)

      retag_omenu = gtk.OptionMenu()
      retag_omenu.show()
      menu = gtk.Menu()
      index = 0
      for choice in self.retag_choices:
         item = gtk.MenuItem(choice)
         item.show()
         menu.append(item)
      retag_omenu.set_menu(menu)

      textview = gtk.TextView()
      textview.show()
      sw = gtk.ScrolledWindow()
      sw.show()
      sw.add_with_viewport(textview)
      dialog.vbox.pack_start(retag_omenu)
      dialog.vbox.pack_start(sw)
      button = gtk.Button("OK")
      button.show()
      button.connect("clicked", self.cb_retag, (event, textview, retag_omenu, dialog))
      retag_omenu.connect("changed", self.cb_changed_retag_omenu, textview)
      dialog.action_area.pack_end(button)
      dialog.show()

   def cb_save_history(self, widget, event, data=None):
      """This callback function is used to save the history.
      to a text file.
      """

      dialog_buttons = (gtk.STOCK_CANCEL
                        , gtk.RESPONSE_CANCEL
                        , gtk.STOCK_SAVE
                        , gtk.RESPONSE_OK)

      file_dialog = gtk.FileChooserDialog(title="Save History"
                                , action=gtk.FILE_CHOOSER_ACTION_SAVE
                                , buttons=dialog_buttons)
      file_dialog.set_current_name('history.txt')
      # Create and add the txt filter
      filter = gtk.FileFilter()
      filter.set_name("Text")
      filter.add_pattern("*." + 'txt')
      file_dialog.add_filter(filter)
      # Create and add the 'all files' filter
      filter = gtk.FileFilter()
      filter.set_name("All files")
      filter.add_pattern("*")
      file_dialog.add_filter(filter)

      result = ""
      if file_dialog.run() == gtk.RESPONSE_OK:
              result = file_dialog.get_filename()
      file_dialog.destroy()

      if result != "":
         o = open(result, 'w');
         o.write("Time,PID,Event,Tag,Value\n") 
         row = self.history_store.get_iter_first()
         while row is not None:
            for x in range(5):
               value = self.history_store.get_value(row,x)
               if value == None:
                  value = ''
               o.write('%s,' % (value))
            o.write('\n')
            old_row = row
            row = self.history_store.iter_children(row)
            if row == None:
               row = self.history_store.iter_next(old_row)
            if row == None:
               row = self.history_store.iter_parent(old_row)
               if row != None:
                  row = self.history_store.iter_next(row)
         o.close()


   def cb_clear_history(self, widget, event, data=None):
      self.history = {}
      self.history_store.clear()

   def cb_size_change(self, widget, event, data=None):
      adj = widget.get_vadjustment()
      adj.set_value( adj.upper - adj.page_size )

   def cb_close_about(self, widget, event, data=None):
      widget.destroy()
      return True

   def cb_close_retag(self, widget, event, data=None):
      widget.destroy()
      return True

   def cb_close_history(self, widget, event, data=None):
      w = self.item_factory.get_item("/View/History")
      w.set_active(False)
      return True

   def cb_close_dialog(self, widget, event, data=None):
      if self.allow_cancel:
         widget.hide()
      return True

   #####################
   #                   #
   #    XML Parsing    #
   #                   #
   #####################

   def _add_device(self, path, device, parent):
      name = device.get("name", "")
      serial = device.get("serial", "")
      location = device.get("location", "")
      did = device.get("id", "-1")
      meta = device.get("meta_device", "")
      d = Device(name, serial, location, did, meta)
      d.path = path+d.name
      parent.add_device(d)

      if d.is_meta_device:
         self.menu_items.append((d.path, None, None, d.id, "<Branch>"))
      else:
         self.menu_items.append((d.path, None, self.cb_mdevice, d.id, "<CheckItem>"))
         self.study.add_actual_device(d)

      for dev in device.findall("Device"):
         p = path + d.name.replace('_', '') + "/"
         self._add_device(p, dev, d)



   def connected_to_middleware(self):
      gobject.source_remove(self.timer)
      self.timer = 0
      self.pbar.hide()
      self.connect_dialog.hide()
      self.allow_cancel = True
      self.participant_dialog.show()
      self.annotator.set_experiment(self.study.name)
      self.annotator.set_dataset(self.testid)
      self.make_inactive()
      
   def show_output(self, event):
      if event.uuid in self.history:
         row = self.history_store.get_iter_first()
         while row is not None:
            if self.history_store.get_value(row, 5).uuid == event.uuid:
               #value = self.history_store.get_value(row, 3) + ";" + event.labelValue
               #self.history_store.set_value(row, 3, value)
               self.history_store.append(row, [None, None, None, event.labelName, event.labelValue, event])
               break
            row = self.history_store.iter_next(row)
      else:
         if event.isTag:
            e_name = event.message
            t_name = event.labelName
            value = event.labelValue
         else:
            e_name = event.command
            t_name = None
            value = event.value
         if self.active:
            psid = self.psid
         else:
            psid = self.testid
         self.history[event.uuid] = event
         self.history_store.append(None, [event.stamp, psid, e_name, t_name, value, event])

   def _clear_items(self, w):
      submenu = w.get_submenu()
      if submenu == None:
         w.set_active(False)
      else:
         for child in submenu.get_children():
            self._clear_items(child)

   def clear_prompt_devices(self):
      w = self.item_factory.get_item("/Device")
      self._clear_items(w)
      self.prompt_devices = []

   def set_prompt_device(self, did):
      self.prompt_device = self.study.devices[did]
      d = self.study.devices[did].get_default_device()
      w = self.item_factory.get_item(d.path.replace('_', ''))
      w.set_active(True)
      
   def load_task(self, exp, t, tooltips):
      vbox = gtk.VBox()
      vbox.show()
      label = gtk.Label(t.name)
      label.show()
      vbox.pack_start(label, False)
      if t.desc != "":
         tooltips.set_tip(label, t.desc)
      if t.start:
         button = gtk.Button("START")
         button.show()
         button.modify_bg(gtk.STATE_NORMAL, self.green)
         button.set_size_request(20,45)
         button.connect("clicked", self.cb_startstop_task, (exp, t))
         vbox.pack_start(button, False)

      for event in t.sorted_events():
         e = t.events[event]
         if e.type == Event.BUTTON:
            button = gtk.Button(e.name)
            button.show()
            button.connect("clicked", self.cb_event_button, (exp, t, e, ""))
            self.button_set.add(button)
            row = button
         elif e.type == Event.FIELD:
            row = gtk.HBox()
            row.show()
            fields = []
            for field in range(e.inputs):
               f = gtk.Entry()
               f.show()
               fields.append(f)
            button = gtk.Button(e.name)
            button.show()
            button.connect("clicked", self.cb_event_button, (exp, t, e, fields))
            self.button_set.add(button)
            for field in fields:
               row.pack_start(field, False)
            row.pack_start(button)
         elif e.type == Event.CHOICE:
            row = gtk.HBox()
            row.show()
            choice_omenus = []
            for cmenu in range(e.inputs):
               choice_omenu = gtk.OptionMenu()
               choice_omenu.show()
               menu = gtk.Menu()
               menu.show()
               for choice in e.choices:
                  item = gtk.MenuItem(choice)
                  item.show()
                  menu.append(item)
               choice_omenu.set_menu(menu)
               choice_omenus.append(choice_omenu)
            button = gtk.Button(e.name)
            button.show()
            button.connect("clicked", self.cb_event_button, (exp, t, e, choice_omenus))
            self.button_set.add(button)
            for choice_omenu in choice_omenus:
               row.pack_start(choice_omenu, False)
            row.pack_start(button)

         if e.desc != "":
            tooltips.set_tip(row, e.desc)
         vbox.pack_start(row, False)
         if e.prompt is not None:
            button.connect("clicked", self.cb_update_prompt, e.prompt)

      row = gtk.HBox()
      row.show()
      field = gtk.Entry()
      field.show()
      button = gtk.Button("Other")
      button.show()
      button.connect("clicked", self.cb_event_button, (exp, t, Event("Other", -1, "", "FIELD"), [field]))
      self.button_set.add(button)
      row.pack_start(field)
      row.pack_start(button, False)
      vbox.pack_end(row, False)
      return vbox

   def load_experiment(self, eid):
      self.eid = eid
      has_prompts = False
      self.button_set.clear()
      self.prompt = None
      # Clear old experiment
      self.tooltips = gtk.Tooltips()
      old = self.global_start_area.get_children()
      for o in old:
         o.destroy()
      old = self.task_list_area.get_children()
      for o in old:
         o.destroy()
      old = self.task_sw.get_child()
      if old is not None:
         old.destroy()
      old = self.prompt_area.get_children()
      for o in old:
         o.destroy()
      # Open new experiment
      exp = self.study.experiments[eid]
      if exp.start == True:
         button = gtk.Button("START")
         button.show()
         button.modify_bg(gtk.STATE_NORMAL, self.green)
         button.set_size_request(20,45)
         button.connect("clicked", self.cb_startstop_task, (exp, None))
         self.global_start_area.pack_start(button, False)
      if exp.questionnaire != None:
         self.questionnaire_button.show()
         if self.questionnaire_buttonh != None:
            self.questionnaire_button.disconnect(self.questionnaire_buttonh)
         self.questionnaire_buttonh = self.questionnaire_button.connect("clicked",
                                                       self.cb_questionnaire_button,                                                                      eid)
      else:
         self.questionnaire_button.hide()
         if self.questionnaire_buttonh != None:
            self.questionnaire_button.disconnect(self.questionnaire_buttonh)
            self.questionnaire_buttonh = None
      # Adjust by Conditions
      if self.condition.has_experiment(eid):
         tasks = []
         prompts = []
         for task in self.condition.get_experiment_tasks(eid):
            tasks.append(task[0])
            try:
               exp.tasks[task[0]].prompt.level = int(task[1])
            except AttributeError,e:
               pass
         for prompt in self.condition.get_experiment_prompts(eid):
            prompts.append(prompt[0])
            exp.prompts[prompt[0]].level = int(prompt[1])
         num_tasks = len(self.condition.experiment_tasks[eid])
         tasks_copy = tasks
      else:
         prompts = []
         for prompt in exp.sorted_prompts():
            exp.prompts[prompt].level = 0
            prompts.append(prompt)
         tasks = exp.sorted_tasks()
         tasks_copy = exp.sorted_tasks()
         num_tasks = len(exp.tasks)
      for t in tasks_copy:
         if exp.tasks[t].prompt != None and exp.tasks[t].prompt.id not in prompts:
            prompts.append(exp.tasks[t].prompt.id)
      # Create Prompts
      prompt_table = gtk.Table(len(prompts), 5, False)
      prompt_table.show()
      prompt_row = 0
      for p in prompts:
         prompt = exp.prompts[p]
         prompt.played = False
         has_prompts = True
         label = gtk.Label(prompt.name)
         label.show()
         #noPromptLabel = gtk.Label("N/A")
         #noPromptLabel.show()
         b1 = gtk.Button("1")
         b1.show()
         b1.connect("clicked", self.cb_prompt, ("1", prompt))
         b2 = gtk.Button("2")
         b2.show()
         b2.connect("clicked", self.cb_prompt, ("2", prompt))
         b3 = gtk.Button("3")
         b3.show()
         b3.connect("clicked", self.cb_prompt, ("3", prompt))
         prompt.buttons = [b1, b2, b3]
         index = prompt.level-1
         if index > -1:
            prompt.buttons[index].modify_bg(gtk.STATE_NORMAL, self.lgreen)
         prompt_table.attach(label, 0, 1, prompt_row, prompt_row+1)
         #prompt_table.attach(noPromptLabel, 1, 2, prompt_row, prompt_row+1)
         prompt_table.attach(b1, 2, 3, prompt_row, prompt_row+1)
         prompt_table.attach(b2, 3, 4, prompt_row, prompt_row+1)
         prompt_table.attach(b3, 4, 5, prompt_row, prompt_row+1)
         prompt_row += 1
      self.prompt_area.pack_start(prompt_table, False)

      if exp.display_mode == Experiment.LIST:
         hbox = gtk.HBox(True,3)
         hbox.show()
         for task in tasks:
            t = exp.tasks[task]
            vbox = self.load_task(exp, t, self.tooltips)
            hbox.pack_start(vbox, False)
            if num_tasks > 1:
               button = gtk.Button(t.name)
               button.show()
               button.connect("clicked", self.cb_task_button, vbox)
               self.button_set.add(button)
               self.task_list_area.pack_start(button)
               self.task_list_sw.show()
            else:
               self.task_list_sw.hide()

         self.task_sw.add_with_viewport(hbox)   
      elif exp.display_mode == Experiment.GRID:
         table = gtk.Table(exp.rows, exp.cols)
         table.show()
         row_n = 0
         col_m = 0
         for task in tasks:
            t = exp.tasks[task]
            vbox = self.load_task(exp, t, self.tooltips)
            table.attach(vbox, col_m, col_m+1, row_n, row_n+1)

            col_m += 1
            if col_m >= exp.cols:
               row_n += 1
               col_m = 0

         self.task_sw.add_with_viewport(table)   
         self.task_list_sw.hide()
      elif exp.display_mode == Experiment.GROUPED:
         hbox = gtk.HBox(True,3)
         hbox.show()
         groups = {}
         for task in tasks:
            t = exp.tasks[task]
            vbox = self.load_task(exp, t, self.tooltips)
            if t.group not in groups:
               group_box = gtk.HBox(True, 3)
               group_box.show()
               groups[t.group] = group_box
               hbox.pack_start(group_box, False)
               button = gtk.Button(t.name)
               button.show()
               button.connect("clicked", self.cb_task_button, group_box)
               self.button_set.add(button)
               self.task_list_area.pack_start(button)
            else:
               group_box = groups[t.group]
            group_box.pack_start(vbox, False)

         self.task_sw.add_with_viewport(hbox)   
         self.task_list_sw.show()
      else:
         print "Other Display Mode"
      if has_prompts:
         self.prompt_sw.show()
      else:
         self.prompt_sw.hide()
      if self.active:
         self.make_active()
      else:
         self.make_inactive()

   def show_questionnaire(self, eid):
      """
      Creates and shows the questionnaire dialog box for the given experiment
      """
      exp = self.study.experiments[eid]
      self.questionnaire_tooltips = gtk.Tooltips()
      t = exp.questionnaire
      if t == None:
         return

      vbox = self.load_task(exp, t, self.questionnaire_tooltips)

      questionnaire_dialog = gtk.Dialog(t.name, self.window)
      questionnaire_dialog.vbox.pack_start(vbox)
      questionnaire_dialog.show()

   def show_error(self, msg, pid=None):
      """
      Displays critical error message
      """
      window = gtk.Window(gtk.WINDOW_TOPLEVEL)
      window.set_title(FULL_NAME)
      window.connect("delete_event", self.cb_quit)
      
      main_box = gtk.VBox(False, 1)
      main_box.show()
      for text in msg:
         label = gtk.Label(text)
         label.show()
         main_box.pack_start(label)
      button = gtk.Button("Quit")
      button.show()
      button.connect('clicked', self.cb_quit, None)
      main_box.pack_start(button)      
      if pid != None:
         button = gtk.Button("Quit All")
         button.show()
         button.connect('clicked', self.cb_quit, None, pid)
         main_box.pack_start(button)      
      window.add(main_box)

      window.show()

   def move_task(self, task_box):
      """
      Moves the given task display to the first position in the box
      """
      self.task_sw.get_child().get_child().reorder_child(task_box, 0)  

   def load_condition(self, condition):
      random.seed(self.seed)
      name = condition.get("name", "")
      cid = int(condition.get("id", "-1"))
      c = Condition(name, cid)
      samples = {}
      for sam in condition.findall("Sample"):
         sid = int(sam.get("id", "-1"))
         choice_list = sam.get("choice_list", "").split(",")
         replace = sam.get("replace", "True")
         if replace.lower() == 'false':
            replace = False
         else:
            replace = True
         s = Sampler(choice_list, replace, sid)
         c.add_sampler(s)
      if self.condition == None:
         self.condition = c
      for exp in condition.findall("Experiment"):
         eid = int(exp.get("id", "-1"))
         tlist = {}
         for task in exp.findall("Task"):
            tid = task.get("id")
            if tid.startswith("*"):
               index = int(tid[1:])
               tid = int(c.get_sample(index))
            else:
               tid = int(tid)
            order = task.get("order")
            if order.startswith("*"):
               index = int(order[1:])
               order = int(c.get_sample(index))
            else:
               order = int(order)
            level = task.get("level", "0")
            if level.startswith("*"):
               index = int(level[1:])
               level = int(c.get_sample(index))
            else:
               level = int(level)
            tlist[order] = (tid, level)
         plist = {}
         for prompt in exp.findall("Prompt"):
            pid = int(prompt.get("id"))
            order = prompt.get("order")
            if order.startswith("*"):
               index = int(order[1:])
               order = int(c.get_sample(index))
            else:
               order = int(order)
            level = prompt.get("level")
            if level.startswith("*"):
               index = int(level[1:])
               level = int(c.get_sample(index))
            else:
               level = int(level)
            plist[order] = (pid, level)
         c.add_experiment(eid, tlist, plist)
      self.study.add_condition(c)


   def make_inactive(self):
      """
      Change the buttons to red indicating test mode has been entered
      """
      for button in self.button_set:
         button.modify_bg(gtk.STATE_NORMAL, self.red)
      self.active = False

   def make_active(self):
      """
      Change the buttons to the default color
      """
      for button in self.button_set:
         button.modify_bg(gtk.STATE_NORMAL, self.default)
      self.active = True

   def set_psid_used(self, psid, condition, device, seed):
      """
      Saves psid in psid file if it does not already exist 
      """
      if psid in self.psid_list:
         return
      else:
         self.psid_list[psid] = (condition.id, device.id, seed)
         f = open("%s.sid" % (self.study.name), "a")
         f.write("%s,%i,%i,%s\n" % (psid, condition.id, device.id, repr(seed)))
         f.close()

   def get_psid_list(self):
      """
      Retrieves a list of all used psids
      """
      f = open("%s.sid" % (self.study.name), "a+")
      f.seek(0)
      psids = OrderedDict()
      for line in f:
         line = line.strip()
         values = line.split(",")
         try:
            """
            values[0] is psid
            values[1] is experimental condition
            values[2] is prompt device id
            values[3] is random number seed
            """
            p = values[0]
            try:
               c = int(values[1])
            except:
               c = 0
            try:
               d = int(values[2])
            except:
               d = 0
            try:
               s = float(values[3])
            except:
               s = self.seed_generator.random()
            psids[p] = (c, d, s)
         except:
            pass
      f.close()
      return psids

   def get_main_menu(self, window):
      """
      Creates the main menu
      """
      accel_group = gtk.AccelGroup()
      item_factory = gtk.ItemFactory(gtk.MenuBar, "<main>", accel_group)
      item_factory.create_items(self.menu_items)
      val, mod = gtk.accelerator_parse("F2")
      accel_group.connect_group(val, mod, 0, self.cb_quick_prompt)
      window.add_accel_group(accel_group)
      self.item_factory = item_factory
      return item_factory.get_widget("<main>")


   def __init__(self, options):
      #Initialize variables
      self.allow_cancel = False
      self.psid = "TEST"
      self.testid = "TEST"
      self.pdata = None
      self.annotator = None
      self.timer = 0
      self.red = gtk.gdk.Color(52428, 0, 0)
      self.green = gtk.gdk.Color(0, 52428, 0)
      self.lgreen = gtk.gdk.Color(43690, 52428, 43690)
      self.dgreen = gtk.gdk.Color(0, 43690, 0)
      self.yellow = gtk.gdk.Color(52428, 52428, 0)
      self.gray = gtk.gdk.Color(52428, 52428, 43690)
      self.white = gtk.gdk.Color(65535, 65535, 65535)
      self.default = gtk.Button().get_style().bg_gc[gtk.STATE_NORMAL]
      self.config = ConfigParser.ConfigParser()
      self.active = False
      self.button_set = set()
      self.prompt_devices = []
      self.prompt = None
      self.questionnaire_tooltips = None
      self.tooltips = None
      self.condition = None
      self.prompt_device = None
      self.xml_tree = None
      self.psid_list = {}
      self.retag_choices = ['Other', 'Ignore', 'Timing', 'New Value']
      self.seed_generator = random.Random()
      self.seed = self.seed_generator.random()

      #  Ensure only one instance ever runs
      lock_file = '.rat.lock'
      self.fp = open('.rat.lock', 'w')
      try:
         fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
         #We are the only instance save pid
         fp = open('.rat.pid', 'w')
         fp.write("%d\n" % (os.getpid()))
         fp.close()
      except IOError:
         # another instance is already running
         fp = open('.rat.pid')
         pid = int(fp.readline().strip())
         fp.close()
         msg = []
         msg.append("Another instance of RAT is currently running")
         msg.append("To prevent data corruption RAT is limited to one instance")
         msg.append("Choosing quit all will end all instances of the RAT and ")
         msg.append("you can then restart the RAT safely.")
         self.show_error(msg, pid)
         return
      except NameError:
         print "Warning unable to verify no other instance of RAT are running"

      #Initialize xmpp configuration
      self.config.add_section("xmpp")
      self.config.set("xmpp", "jid", "username@server")
      self.config.set("xmpp", "pass", "password")
      self.config.set("xmpp", "manager", "")
      self.config.read(str(options.config))
      username = self.config.get("xmpp", "jid")
      password = self.config.get("xmpp", "pass")
      manager = self.config.get("xmpp", "manager")

      # Create the main menu.
      self.menu_items = [
         ("/_File",                    None,         None, 0, "<Branch>"),
         ("/File/_Save History", "<control>S", self.cb_save_history, 0, None),
         ("/File/_Clear History",      None, self.cb_clear_history, 0, None),
         ("/File/sep1",                None,         None, 0, "<Separator>"),
         ("/File/_Stop Participant",   None, self.cb_stop_participant, 0, None),
         ("/File/_Change Participant", "<control>N", self.cb_mchange, 0, None),
         ("/File/sep2",                None,         None, 0, "<Separator>"),
         #("/File/_Change XMPP Connection", None,   self.cb_reconnect, 0, None),
         ("/File/Quit",                "<control>Q", self.cb_quit, 0, None),
         ("/_View",                    None,         None, 0, "<Branch>"),
         ("/View/_History",            "<control>H", self.cb_mhistory, 0, "<CheckItem>"),
         ("/_Experiment",              None,         None, 0, "<Branch>"),
         ("/_Device",                  None,         None, 0, "<Branch>"),
         ("/_Help",                    None,         None, 0, "<Branch>"),
         ("/_Help/_About",             None,         self.cb_mabout, 0, None),
      ]
 
      # Read the xml config file
      try:
         self.xml_tree = ET.parse(options.xml)
      except Exception, e:
         msg = []
         msg.append("I was unable to read the supplied config file")
         msg.append("While trying I encountered the following error")
         msg.append(str(e))
         self.show_error(msg)
         return
      root = self.xml_tree.getroot()
      name = root.get("name", "")
      self.study = Study(name)
      self.testid = root.get("default_id", "TEST")
      self.psid = root.get("default_id", "TEST")
      # Read Devices
      for device in root.findall("Device"):
         self._add_device("/Device/", device, self.study)
      # Read Prompt
      for prompt in root.findall("Prompt"):
         p = Prompt(prompt.get("name", ""), prompt.get("id", "-1"), prompt.get("file", ""), prompt.get("default_device", ""))
         self.study.add_prompt(p)
      # Read Experiment
      for exp in root.findall("Experiment"):
         e = Experiment(exp.get("name", ""), exp.get("id", "-1"), exp.get("display_mode", "list"))
         start = exp.get("start", "")
         if start != "" and start.lower() != "false":
            e.start = True
         if e.display_mode != Experiment.LIST:
            e.set_columns(exp.get("columns", "1"))
         for prompt in exp.findall("Prompt"):
            pid = int(prompt.get("id", "-1"))
            if pid not in self.study.prompts:
               self.study.add_prompt(Prompt(prompt.get(name, ""), pid, prompt.get("file", ""), prompt.get("default_device", "")))
            e.prompt = self.study.prompts[pid]
            e.add_prompt(e.prompt)
         # Read questionnaire
         questionnaire = exp.find("Questionnaire")
         if questionnaire != None:
            name = questionnaire.get("name", "")
            tid = int(questionnaire.get("id", "-1"))
            desc = questionnaire.get("desc", "")
            group = questionnaire.get("group", "1")
            t = Task(name, tid, desc, group)
            for prompt in questionnaire.findall("Prompt"):
               pid = int(prompt.get("id", "-1"))
               if pid not in self.study.prompts:
                  self.study.add_prompt(Prompt(prompt.get(name, ""), pid, prompt.get("file", ""), prompt.get("default_device", "")))
               t.prompt = self.study.prompts[pid]
               e.add_prompt(t.prompt)
            for event in questionnaire.findall("Event"):
               name = event.get("name", "")
               eid = int(event.get("id", "-1"))
               desc = event.get("desc", "")
               etype = event.get("type", "")
               einputs = int(event.get("inputs", "1"))
               ev = Event(name, eid, desc, etype, einputs)
               if ev.type == Event.CHOICE:
                  choice_list = event.get("choice_list", "")
                  ev.set_choices(choice_list.split(","))
               for prompt in event.findall("Prompt"):
                  pid = int(prompt.get("id", "-1"))
                  if pid not in self.study.prompts:
                     self.study.add_prompt(Prompt(prompt.get(name, ""), pid, prompt.get("file", ""), prompt.get("default_device", "")))
                  ev.prompt = self.study.prompts[pid]
                  e.add_prompt(ev.prompt)
               t.add_event(ev)
            e.questionnaire = t
         # Read Task   
         for task in exp.findall("Task"):
            name = task.get("name", "")
            tid = int(task.get("id", "-1"))
            desc = task.get("desc", "")
            group = task.get("group", "1")
            t = Task(name, tid, desc, group)
            start = task.get("start", "")
            if start.lower() != "false":
               t.start = True
            for prompt in task.findall("Prompt"):
               pid = int(prompt.get("id", "-1"))
               if pid not in self.study.prompts:
                  self.study.add_prompt(Prompt(prompt.get(name, ""), pid, prompt.get("file", ""), prompt.get("default_device", "")))
               t.prompt = self.study.prompts[pid]
               e.add_prompt(t.prompt)
            for event in task.findall("Event"):
               name = event.get("name", "")
               eid = int(event.get("id", "-1"))
               desc = event.get("desc", "")
               etype = event.get("type", "")
               einputs = int(event.get("inputs", "1"))
               ev = Event(name, eid, desc, etype, einputs)
               if ev.type == Event.CHOICE:
                  choice_list = event.get("choice_list", "")
                  ev.set_choices(choice_list.split(","))
               for prompt in event.findall("Prompt"):
                  pid = int(prompt.get("id", "-1"))
                  if pid not in self.study.prompts:
                     self.study.add_prompt(Prompt(prompt.get(name, ""), pid, prompt.get("file", ""), prompt.get("default_device", "")))
                  ev.prompt = self.study.prompts[pid]
                  e.add_prompt(ev.prompt)
               t.add_event(ev)
            e.add_task(t)
         self.study.add_experiment(e)
      # Read the Conditions
      for condition in root.findall("Condition"):
         self.load_condition(condition)

      group = None
      for key in self.study.sorted_experiments():
         e = self.study.experiments[key]
         path = "/Experiment/%s" % (e.name)
         if group == None:
            group = path
            self.menu_items.append((path, None, self.cb_mexperiment, e.id, "<RadioItem>"))
         else:
            self.menu_items.append((path, None, self.cb_mexperiment, e.id, group))

      # Create the main window
      self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
      self.window.maximize()
      self.window.set_title(FULL_NAME + ": " + self.study.name)
      self.window.connect("delete_event", self.cb_quit)
      
      main_box = gtk.VBox(False, 1)
      main_box.show()
      self.window.add(main_box)

      menubar = self.get_main_menu(self.window)
      menubar.show()
      device = root.find("Device")
      did = int(device.get("id", "-1"))
      self.set_prompt_device(did)
      info_box = gtk.HBox(spacing=10)
      info_box.show()
      hbox = gtk.HBox()
      hbox.show()

      label = gtk.Label("Participant ID:")
      label.show()
      self.psid_label = gtk.Label(self.testid)
      self.psid_label.show()
      self.questionnaire_button = gtk.Button("Questionnaire")
      self.questionnaire_buttonh = None
      self.start_button = gtk.Button("Start Participant: " + self.psid)
      self.start_button.show()
      self.start_button.modify_bg(gtk.STATE_NORMAL, self.green)
      self.start_button.connect("clicked", self.cb_start_participant, None)
      animation = gtk.gdk.PixbufAnimation("images/run_icon.gif")
      self.run_icon = gtk.Image()
      self.run_icon.set_from_animation(animation)
      self.stop_icon = gtk.Image()
      self.stop_icon.show()
      self.stop_icon.set_from_file("images/stop_icon.png")
      info_box.pack_start(label, False)
      info_box.pack_start(self.psid_label, False)
      info_box.pack_start(self.questionnaire_button, False)
      info_box.pack_start(self.start_button)
      info_box.pack_end(self.stop_icon, False)
      info_box.pack_end(self.run_icon, False)

      self.prompt_sw = gtk.ScrolledWindow()
      self.prompt_sw.show()
      self.prompt_sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
      self.prompt_area = gtk.VBox()
      self.prompt_area.show()
      self.prompt_sw.add_with_viewport(self.prompt_area)

      vbox = gtk.VBox()
      vbox.show()
      self.global_start_area = gtk.HBox(True)
      self.global_start_area.show()
      self.task_sw = gtk.ScrolledWindow()
      self.task_sw.show()
      self.task_list_area = gtk.HBox(True)
      self.task_list_area.show()
      self.task_list_sw = gtk.ScrolledWindow()
      self.task_list_sw.show()
      self.task_list_sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_NEVER)
      self.task_list_sw.add_with_viewport(self.task_list_area)
      vbox.pack_start(self.global_start_area, False)
      vbox.pack_start(self.task_list_sw, False)
      vbox.pack_start(self.task_sw, True)
      
      hbox.pack_start(self.prompt_sw, False)
      hbox.pack_start(vbox)
      main_box.pack_start(menubar, False)
      main_box.pack_start(info_box, False)
      main_box.pack_start(hbox)

      self.load_experiment(self.study.first_experiment())

      # Build history dialog
      self.history_store = gtk.TreeStore(str, str, str, str, str, gobject.TYPE_PYOBJECT)
      self.history_tv = gtk.TreeView(self.history_store)
      self.history_tv.show()
      self.history_tv.connect("row-activated", self.cb_history_row_activated)
      self.history_tv.connect("size-allocate", self.cb_size_change)
      self.history_tv.set_rules_hint(True)

      rendererText = gtk.CellRendererText()
      column = gtk.TreeViewColumn("Time", rendererText, text=0)
      column.set_sort_column_id(0)    
      column.set_resizable(True)
      self.history_tv.append_column(column)

      rendererText = gtk.CellRendererText()
      column = gtk.TreeViewColumn("PID", rendererText, text=1)
      column.set_sort_column_id(1)
      column.set_resizable(True)
      self.history_tv.append_column(column)

      rendererText = gtk.CellRendererText()
      column = gtk.TreeViewColumn("Event", rendererText, text=2)
      column.set_sort_column_id(2)
      column.set_resizable(True)
      self.history_tv.append_column(column) 

      rendererText = gtk.CellRendererText()
      column = gtk.TreeViewColumn("Tag", rendererText, text=3)
      column.set_sort_column_id(3)
      column.set_resizable(True)
      self.history_tv.append_column(column) 

      rendererText = gtk.CellRendererText()
      column = gtk.TreeViewColumn("Value", rendererText, text=4)
      column.set_sort_column_id(3)
      column.set_resizable(True)
      self.history_tv.append_column(column) 

      self.history = {}
      history_sw = gtk.ScrolledWindow()
      history_sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
      history_sw.show()
      history_sw.add(self.history_tv)

      self.history_dialog = gtk.Dialog("History", self.window)
      self.history_dialog.vbox.pack_start(history_sw)
      self.history_dialog.set_default_size(600, 350)
      self.history_dialog.connect("delete_event", self.cb_close_history)

      # Build connect dialog and components
      self.cd_userfield = gtk.Entry()
      self.cd_passfield = gtk.Entry()
      self.cd_managerfield = gtk.Entry()
      self.cd_local = gtk.CheckButton("Use Local Emulator")

      self.cd_userfield.set_text(username)
      self.cd_passfield.set_text(password)
      self.cd_managerfield.set_text(manager)
      table = gtk.Table(6, 2, False)
 
      label = gtk.Label("Username")
      table.attach(label, 0, 1, 0, 1)
      table.attach(self.cd_userfield, 1, 2, 0, 1)
      label.show()
      self.cd_userfield.show()

      label = gtk.Label("Password")
      table.attach(label, 0, 1, 1, 2)
      table.attach(self.cd_passfield, 1, 2, 1, 2)
      label.show()
      self.cd_passfield.show()

      label = gtk.Label("Manager")
      table.attach(label, 0, 1, 2, 3)
      table.attach(self.cd_managerfield, 1, 2, 2, 3)
      label.show()
      self.cd_managerfield.show()

      table.attach(self.cd_local, 0, 2, 3, 4)
      self.cd_local.show()

      self.pbar = gtk.ProgressBar()
      self.timer = gobject.timeout_add (100, progress_timeout, self.pbar)
      table.attach(self.pbar, 0, 2, 4, 5)
      table.show()
      
      button = gtk.Button('Connect')
      button.connect('clicked', self.cb_connect_to_middleware, None)
      button.show()

      qButton = gtk.Button('Quit')
      qButton.connect('clicked', self.cb_quit, None)
      qButton.show()

      self.connect_dialog = gtk.Dialog("Connect to Server", self.window, gtk.DIALOG_MODAL)
      self.connect_dialog.action_area.pack_start(qButton)
      self.connect_dialog.action_area.pack_start(button)
      self.connect_dialog.vbox.pack_start(table)
      self.connect_dialog.connect("delete_event", self.cb_close_dialog)

      # Build particpant dialog and components
      self.psid_list = self.get_psid_list()

      self.pd_new_rbutton = gtk.RadioButton(label="New")
      self.pd_new_rbutton.show()
      self.pd_prompt_omenu = gtk.OptionMenu()
      self.pd_prompt_omenu.show()
      menu = gtk.Menu()
      index = 0
      self.pd_prompt_omenu_translation = {}
      for key in self.study.sorted_devices():
         item = gtk.MenuItem(self.study.devices[key].name)
         item.show()
         menu.append(item)
         self.pd_prompt_omenu_translation[index] = key
         index += 1
      self.pd_prompt_omenu.set_menu(menu)
      self.pd_prompt_omenu.connect("changed", self.cb_pd_new_entry, None)

      self.pd_condition_omenu = gtk.OptionMenu()
      self.pd_condition_omenu.show()
      menu = gtk.Menu()
      index = 0
      self.pd_condition_omenu_translation = {}
      for key in self.study.sorted_conditions():
         item = gtk.MenuItem(self.study.conditions[key].name)
         item.show()
         menu.append(item)
         self.pd_condition_omenu_translation[index] = key
         index += 1
      self.pd_condition_omenu.set_menu(menu)
      self.pd_condition_omenu.connect("changed", self.cb_pd_new_entry, None)
             
      self.pd_age = gtk.Entry()
      self.pd_age.show()
      self.pd_age.connect("focus-in-event", self.cb_pd_new_entry, None)
      self.pd_gender = gtk.Entry()
      self.pd_gender.show()
      self.pd_gender.connect("focus-in-event", self.cb_pd_new_entry, None)
      self.pd_edu = gtk.Entry()
      self.pd_edu.show()
      self.pd_edu.connect("focus-in-event", self.cb_pd_new_entry, None)
      self.pd_hand = gtk.Entry()
      self.pd_hand.show()
      self.pd_hand.connect("focus-in-event", self.cb_pd_new_entry, None)
      self.pd_existing_rbutton = gtk.RadioButton(self.pd_new_rbutton, "Existing")
      self.pd_existing_rbutton.show()
      self.pd_test_rbutton = gtk.RadioButton(self.pd_new_rbutton, "Test Mode")
      self.pd_new_entry = gtk.Entry()
      self.pd_new_entry.show()
      self.pd_new_entry.connect("activate", self.cb_set_participant, None)
      self.pd_new_entry.connect("focus-in-event", self.cb_pd_new_entry, None)
      self.pd_existing_omenu = gtk.OptionMenu()
      self.pd_existing_omenu.show()
      menu = gtk.Menu()
      for psid in self.psid_list:
         item = gtk.MenuItem(psid)
         item.show()
         menu.append(item)
      self.pd_existing_omenu.set_menu(menu)
      self.pd_existing_omenu.connect("changed", self.cb_pd_existing_omenu, None)

      table = gtk.Table(11, 3, False)

      table.attach(self.pd_new_rbutton, 0, 1, 0, 1)
      table.attach(self.pd_new_entry, 1, 3, 0, 1)
      label = gtk.Label("Prompt")
      label.show()
      label.set_alignment(xalign=0.0, yalign=0.5)
      table.attach(label, 1, 2, 1, 2)
      table.attach(self.pd_prompt_omenu, 2, 3, 1, 2)
      label = gtk.Label("6 Act. Condition")
      label.show()
      label.set_alignment(xalign=0.0, yalign=0.5)
      table.attach(label, 1, 2, 2, 3)
      table.attach(self.pd_condition_omenu, 2, 3, 2, 3)
      sep = gtk.HSeparator()
      sep.show()
      table.attach(sep, 0, 3, 7, 8, ypadding=4)


      table.attach(self.pd_existing_rbutton, 0, 1, 8, 9)
      table.attach(self.pd_existing_omenu, 1, 3, 8, 9)
      sep = gtk.HSeparator()
      sep.show()
      table.attach(sep, 0, 3, 9, 10, ypadding=4)

      table.attach(self.pd_test_rbutton, 0, 1, 10, 11)
      self.pd_test_rbutton.show()
      table.show()

      self.pd_ok_button = gtk.Button("OK")
      self.pd_ok_button.show()
      self.pd_button_h = self.pd_ok_button.connect('clicked', self.cb_set_participant, self.psid_list.keys())

      self.participant_dialog = gtk.Dialog("Open Participant", self.window, gtk.DIALOG_MODAL)
      self.participant_dialog.action_area.pack_end(self.pd_ok_button)
      self.participant_dialog.vbox.pack_start(table)
      self.participant_dialog.connect("delete_event", self.cb_close_dialog)
      

      self.window.show()
      self.connect_dialog.show()

class Tag:
   """
   Label,Value pair used for tagging events
   """

   def __init__(self):
      self.label = ""
      self.value = ""

class Study:
   """
   Class which organizes the study into experiments, tasks, etc.
   """

   def __init__(self, name):
      self.name = name
      self.devices = {}
      self.actual_devices = {}
      self.prompts = {}
      self.experiments = {}
      self.conditions = {}

   def add_device(self, device):
      self.devices[device.id] = device

   def add_actual_device(self, device):
      self.actual_devices[device.id] = device

   def add_prompt(self, prompt):
      self.prompts[prompt.id] = prompt

   def add_experiment(self, experiment):
      self.experiments[experiment.id] = experiment

   def add_condition(self, condition):
      self.conditions[condition.id] = condition

   def sorted_devices(self):
      for key in sorted(self.devices.iterkeys()):
         yield key

   def sorted_prompts(self):
      for key in sorted(self.prompts.iterkeys()):
         yield key

   def sorted_experiments(self):
      for key in sorted(self.experiments.iterkeys()):
         yield key

   def sorted_conditions(self):
      for key in sorted(self.conditions.iterkeys()):
         yield key

   def first_experiment(self):
      return min(self.experiments)

class Device:
   
   def __init__(self, name, serial, location, did, meta):
      self.name = name
      self.serial = serial
      self.location = location
      self.id = int(did)
      self.default_id = 999999
      self.is_meta_device = bool(meta)
      self.devices = {}

   def add_device(self, device):
      if device.id < self.default_id:
         self.default_id = device.id
      self.devices[device.id] = device

   def get_default_device(self):
      if len(self.devices) > 0:
         return self.devices[self.default_id].get_default_device()
      else:
         return self
      

   def sorted_devices(self):
      for key in sorted(self.devices.iterkeys()):
         yield key

class Prompt:

   def __init__(self, name, pid, filename, devices):
      self.name = name
      self.id = int(pid)
      self.filename = filename
      self.default_devices = devices.split(",")
      self.level = 0
      self.played = False

class Sampler:

   def __init__(self, choice_list, replace, sid):
      self.choice_list = choice_list
      self.replace = replace
      self.id = int(sid)
      self.active_choices = self.choice_list

   def get_sample(self):
      if len(self.active_choices) == 0:
         self.active_choices = self.choice_list
      sample = random.choice(self.active_choices)
      if not self.replace:
         self.active_choices.remove(sample)
      return sample


class Condition:

   def __init__(self, name, cid):
      self.name = name
      self.id = int(cid)
      self.experiment_tasks = {}
      self.experiment_prompts = {}
      self.samplers = {}

   def add_experiment(self, eid, tasks, prompts):
      self.experiment_tasks[eid] = tasks
      self.experiment_prompts[eid] = prompts

   def add_sampler(self, sampler):
      self.samplers[sampler.id] = sampler
   
   def get_sample(self, sid):
      return self.samplers[sid].get_sample()

   def has_experiment(self, eid):
      return eid in self.experiment_tasks

   def get_experiment_tasks(self, eid):
      for key in sorted(self.experiment_tasks[eid].iterkeys()):
         yield self.experiment_tasks[eid][key]

   def get_experiment_prompts(self, eid):
      for key in sorted(self.experiment_prompts[eid].iterkeys()):
         yield self.experiment_prompts[eid][key]

class Experiment:
   LIST = 0
   GRID = 1
   GROUPED = 2
   
   def __init__(self, name, eid, mode):
      self.name = name
      self.id = int(eid)
      self.rows = 1
      self.cols = 1 
      if mode == "grid":
         self.display_mode = Experiment.GRID
      elif mode == "grouped":
         self.display_mode = Experiment.GROUPED
      else:
         self.display_mode = Experiment.LIST
      self.start = False
      self.questionnaire = None
      self.prompt = None 
      self.tasks = {}
      self.prompts = {}
      self.buttons = []

   def set_columns(self, cols):
      self.cols = int(cols)

   def add_task(self, task):
      self.tasks[task.id] = task
      if len(self.tasks) / self.cols > self.rows:
         self.rows += 1

   def add_prompt(self, prompt):
      self.prompts[prompt.id] = prompt

   def sorted_tasks(self):
      for key in sorted(self.tasks.iterkeys()):
         yield key

   def sorted_prompts(self):
      for key in sorted(self.prompts.iterkeys()):
         yield key

class Task:

   def __init__(self, name, tid, desc, group):
      self.name = name
      self.id = int(tid)
      self.desc = desc
      self.events = {}
      self.group = group
      self.prompt = None 
      self.start = False

   def add_event(self, event):
      self.events[event.id] = event

   def sorted_events(self):
      for key in sorted(self.events.iterkeys()):
         yield key

class Event:

   BUTTON = "button"
   FIELD = "field"
   CHOICE = "choice"

   def __init__(self, name, eid, desc, type, inputs=1):
      self.name = name
      self.id = int(eid)
      self.desc = desc
      if type == "":
         type = Event.BUTTON
      self.type = type.lower()
      self.prompt = None
      self.choices = []
      self.inputs = inputs

   def set_choices(self, choices):
      self.choices = choices



def main():
   gtk.main()

def sig_handler(sig, frame):
   global a
   a.cb_quit(None, None)

if __name__ == "__main__":
   global a
   signal.signal(signal.SIGINT, sig_handler)
   signal.signal(signal.SIGTERM, sig_handler)
   parser = optparse.OptionParser(usage="usage: %prog [options]")
   parser.add_option("-c",
                   "--config",
                   dest="config",
                   help="Custom annotator config file.",
                   default="Rat.config")
   parser.add_option("-x",
                   "--xml",
                   dest="xml",
                   help="Custom xml file.",
                   default="config.xml")
   (options, args) = parser.parse_args()
   a = RAT_GUI(options)
   main()
